import logging
import sys
import time
import openai

import graphsignal
from graphsignal.endpoint_trace import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.recorder_utils import patch_method, unpatch_method
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class OpenAIRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None
        self._is_sampling = False

    def setup(self):
        if not graphsignal._agent.auto_instrument:
            return

        self._framework = signals_pb2.FrameworkInfo()
        self._framework.type = signals_pb2.FrameworkInfo.FrameworkType.OPENAI_FRAMEWORK
        parse_semver(self._framework.version, openai.version.VERSION)

        if compare_semver(self._framework.version, (0, 26, 0)) < 1:
            logger.debug('OpenAI tracing is only supported for >= 0.26.0.')
            return

        self._instrument(openai.Completion, 'create', 'openai.Completion.create', self.trace_completion)
        self._instrument(openai.Completion, 'acreate', 'openai.Completion.acreate', self.trace_completion)
        self._instrument(openai.Edit, 'create', 'openai.Edit.create', self.trace_edits)
        self._instrument(openai.Edit, 'acreate', 'openai.Edit.acreate', self.trace_edits)
        self._instrument(openai.Embedding, 'create', 'openai.Embedding.create', self.trace_embedding)
        self._instrument(openai.Embedding, 'acreate', 'openai.Embedding.acreate', self.trace_embedding)
        self._instrument(openai.Image, 'create', 'openai.Image.create', self.trace_image)
        self._instrument(openai.Image, 'acreate', 'openai.Image.acreate', self.trace_image)
        self._instrument(openai.Image, 'create_variation', 'openai.Image.create_variation', self.trace_image)
        self._instrument(openai.Image, 'acreate_variation', 'openai.Image.acreate_variation', self.trace_image)
        self._instrument(openai.Image, 'create_edit', 'openai.Image.create_edit', self.trace_image)
        self._instrument(openai.Image, 'acreate_edit', 'openai.Image.acreate_edit', self.trace_image)
        self._instrument(openai.Moderation, 'create', 'openai.Moderation.create', self.trace_moderation)
        self._instrument(openai.Moderation, 'acreate', 'openai.Moderation.acreate', self.trace_moderation)

    def shutdown(self):
        self._uninstrument(openai.Completion, 'create', 'openai.Completion.create')
        self._uninstrument(openai.Completion, 'acreate', 'openai.Completion.acreate')
        self._uninstrument(openai.Edit, 'create', 'openai.Edit.create')
        self._uninstrument(openai.Edit, 'acreate', 'openai.Edit.acreate')
        self._uninstrument(openai.Embedding, 'create', 'openai.Embedding.create')
        self._uninstrument(openai.Embedding, 'acreate', 'openai.Embedding.acreate')
        self._uninstrument(openai.Image, 'create', 'openai.Image.create')
        self._uninstrument(openai.Image, 'acreate', 'openai.Image.acreate')
        self._uninstrument(openai.Image, 'create_variation', 'openai.Image.create_variation')
        self._uninstrument(openai.Image, 'acreate_variation', 'openai.Image.acreate_variation')
        self._uninstrument(openai.Image, 'create_edit', 'openai.Image.create_edit')
        self._uninstrument(openai.Image, 'acreate_edit', 'openai.Image.acreate_edit')
        self._uninstrument(openai.Moderation, 'create', 'openai.Moderation.create')
        self._uninstrument(openai.Moderation, 'acreate', 'openai.Moderation.acreate')

    def _instrument(self, obj, func_name, endpoint, trace_func):
        def before_func(args, kwargs):
            return graphsignal.start_trace(endpoint=endpoint)

        def after_func(args, kwargs, ret, exc, trace):
            trace.measure()
            try:
                if exc is not None:
                    trace.set_exception(exc)

                trace_func(trace, kwargs, ret, exc)
            except Exception as e:
                logger.debug('Error tracing %s', func_name, exc_info=True)

            trace.stop()

        if not patch_method(obj, func_name, before_func=before_func, after_func=after_func):
            logger.debug('Cannot instrument %s.', endpoint)

    def _uninstrument(self, obj, func_name, endpoint):
        if not unpatch_method(obj, func_name):
            logger.debug('Cannot uninstrument %s.', endpoint)

    def trace_completion(self, trace, kwargs, ret, exc):
        if self._is_sampling:
            param_args = [
                'model',
                'max_tokens',
                'temperature',
                'top_p',
                'n',
                'stream',
                'logprobs',
                'echo',
                'stop',
                'presence_penalty',
                'frequency_penalty',
                'best_of'
            ]
            for param_arg in param_args:
                if param_arg in kwargs:
                    trace.set_param(param_arg, kwargs[param_arg])

        if 'stream' in kwargs and kwargs['stream']:
            if 'prompt' in kwargs:
                trace.set_data('prompt', kwargs['prompt'])
            return

        prompt_usage = {}
        completion_usage = {
            'finish_reason_stop': 0,
            'finish_reason_length': 0
        }
        if 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']
            if 'completion_tokens' in ret['usage']:
                completion_usage['token_count'] = ret['usage']['completion_tokens']

        if 'prompt' in kwargs:
            trace.set_data('prompt', kwargs['prompt'], extra_counts=prompt_usage)

        if 'choices' in ret:
            completion = []
            for choice in ret['choices']:
                if 'finish_reason' in choice:
                    if choice['finish_reason'] == 'stop':
                        completion_usage['finish_reason_stop'] += 1
                    elif choice['finish_reason'] == 'length':
                        completion_usage['finish_reason_length'] += 1
                if 'text' in choice:
                    completion.append(choice['text'])
            trace.set_data('completion', completion, extra_counts=completion_usage)

    def trace_edits(self, trace, kwargs, ret, exc):
        if self._is_sampling:
            param_args = [
                'model',
                'temperature',
                'top_p',
                'n'
            ]
            for param_arg in param_args:
                if param_arg in kwargs:
                    trace.set_param(param_arg, kwargs[param_arg])

        prompt_usage = {}
        completion_usage = {}
        if 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']
            if 'completion_tokens' in ret['usage']:
                completion_usage['token_count'] = ret['usage']['completion_tokens']

        if 'input' in kwargs:
            trace.set_data('input', kwargs['input'])

        if 'instruction' in kwargs:
            trace.set_data('instruction', kwargs['instruction'], extra_counts=prompt_usage)

        if 'choices' in ret:
            edits = []
            for choice in ret['choices']:
                if 'text' in choice:
                    edits.append(choice['text'])
            trace.set_data('edits', edits, extra_counts=completion_usage)

    def trace_embedding(self, trace, kwargs, ret, exc):
        if self._is_sampling:
            param_args = [
                'engine'
            ]
            for param_arg in param_args:
                if param_arg in kwargs:
                    trace.set_param(param_arg, kwargs[param_arg])

        prompt_usage = {}
        if 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']

        if 'input' in kwargs:
            trace.set_data('input', kwargs['input'], extra_counts=prompt_usage)

        if 'data' in ret:
            embedding = []
            for choice in ret['data']:
                if 'embedding' in choice:
                    embedding.append(choice['embedding'])
            trace.set_data('embedding', embedding)

    def trace_image(self, trace, kwargs, ret, exc):
        if self._is_sampling:
            param_args = [
                'n',
                'size',
                'response_format'
            ]
            for param_arg in param_args:
                if param_arg in kwargs:
                    trace.set_param(param_arg, kwargs[param_arg])

        if 'prompt' in kwargs:
            trace.set_data('prompt', kwargs['prompt'])

        if 'data' in ret:
            image_data = []
            for image in ret['data']:
                if 'url' in image:
                    image_data.append(image['url'])
                elif 'b64_json' in image:
                    image_data.append(image['b64_json'])
            trace.set_data('image', image_data)

    def trace_moderation(self, trace, kwargs, ret, exc):
        if self._is_sampling:
            param_args = [
                'model'
            ]
            for param_arg in param_args:
                if param_arg in kwargs:
                    trace.set_param(param_arg, kwargs[param_arg])

        if 'input' in kwargs:
            trace.set_data('input', kwargs['input'])

    def on_trace_start(self, signal, context, options):
        self._is_sampling = True

    def on_trace_stop(self, signal, context, options):
        self._is_sampling = False

    def on_trace_read(self, signal, context, options):
        if self._framework:
            signal.frameworks.append(self._framework)




import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.recorders.recorder_utils import patch_method

logger = logging.getLogger('graphsignal')


class Test:
    def __init__(self):
        pass

    def test(self, a, b, c=None):
        return a + 1

    def test_exc(self):
        raise Exception('exc1')


class RecorderUtilsTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_patch_method(self):
        obj = Test()

        before_func_called = False
        def before_func(args, kwargs):
            nonlocal before_func_called
            before_func_called = True
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {'c': 3})
            return dict(d=1)

        after_func_called = False
        def after_func(args, kwargs, ret, exc, context):
            nonlocal after_func_called
            after_func_called = True
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {'c': 3})
            self.assertEqual(ret, 2)
            self.assertIsNone(exc)
            self.assertEqual(context['d'], 1)

        self.assertTrue(patch_method(obj, 'test', before_func=before_func, after_func=after_func))

        obj.test(1, 2, c=3)

        self.assertTrue(before_func_called)
        self.assertTrue(after_func_called)

    def test_patch_method_exc(self):
        obj = Test()

        after_func_called = False
        def after_func(args, kwargs, ret, exc, context):
            nonlocal after_func_called
            after_func_called = True
            self.assertEqual(str(exc), 'exc1')

        self.assertTrue(patch_method(obj, 'test_exc', after_func=after_func))

        with self.assertRaises(Exception) as context:
            obj.test_exc()

        self.assertTrue(after_func_called)


from typing import Optional
import logging
import time
from torch import Tensor
import pytorch_lightning
from pytorch_lightning.callbacks.base import Callback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver, add_framework_param

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, tags: Optional[dict] = None):
        super().__init__()

        self._is_initialized = False
        self._framework = None
        self._model_info = None
        self._rank = None
        self._local_rank = None
        self._node_rank = None
        self._trace = None
        self._tags = tags
        self._model_size_mb = None

    def on_predict_start(self, trainer, pl_module):
        self._configure(trainer, pl_module)

    def on_predict_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_predict_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_trace('predict_batch', batch_idx)

    def on_predict_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._stop_trace()

    def _configure(self, trainer, pl_module):
        try:
            if not self._is_initialized:
                self._is_initialized = True

                self._framework = signals_pb2.FrameworkInfo()
                self._framework.type = signals_pb2.FrameworkInfo.FrameworkType.PYTORCH_LIGHTNING_FRAMEWORK
                parse_semver(self._framework.version, pytorch_lightning.__version__)

                if self._check_param(trainer, 'world_size') and trainer.world_size >= 0:
                    add_framework_param(self._framework, 'world_size', trainer.world_size)
                if self._check_param(trainer, 'node_rank') and trainer.node_rank >= 0:
                    add_framework_param(self._framework, 'node_rank', trainer.node_rank)
                    self._node_rank = trainer.node_rank
                if self._check_param(trainer, 'local_rank') and trainer.local_rank >= 0:
                    add_framework_param(self._framework, 'local_rank', trainer.local_rank)
                    self._local_rank = trainer.local_rank
                if self._check_param(trainer, 'global_rank') and trainer.global_rank >= 0:
                    add_framework_param(self._framework, 'global_rank', trainer.global_rank)
                    self._rank = trainer.global_rank

                self._model_info = signals_pb2.ModelInfo()
                self._model_info.model_format = signals_pb2.ModelInfo.ModelFormat.PYTORCH_FORMAT
                model_size_mb = pytorch_lightning.utilities.memory.get_model_size_mb(pl_module)
                if model_size_mb:
                    self._model_info.model_size_bytes = int(model_size_mb * 1e6)
        except Exception:
            logger.error('Error configuring PyTorch Lightning callback', exc_info=True)

    def _start_trace(self, endpoint, batch_idx):
        if not self._trace:
            self._trace = graphsignal.start_trace(
                endpoint=endpoint,
                tags=self._tags)
            self._trace.set_tag('batch', batch_idx)

    def _stop_trace(self):
        if self._trace:
            if self._trace.is_sampling():
                self._update_signal()
            self._trace.stop()
            self._trace = None

    def _update_signal(self):
        try:
            signal = self._trace._signal

            if self._framework:
                signal.frameworks.append(self._framework)
            if self._model_info:
                signal.model_info.CopyFrom(self._model_info)
            if self._rank is not None:
                signal.process_usage.rank = self._rank
                signal.process_usage.has_rank = True
            if self._local_rank is not None:
                signal.process_usage.local_rank = self._local_rank
                signal.process_usage.has_local_rank = True
            if self._node_rank is not None:
                signal.node_usage.node_rank = self._node_rank
                signal.node_usage.has_node_rank = True

        except Exception as exc:
            logger.debug('Error in Hugging Face callback', exc_info=True)
            self._trace._add_agent_exception(exc)

    def _check_param(self, trainer, param):
        value = getattr(trainer, param, None)
        return isinstance(value, (str, int, float, bool))


from typing import Optional
import logging

from tensorflow import keras
from tensorflow.keras.callbacks import Callback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, tags: Optional[dict] = None):
        super().__init__()

        self._is_initialized = False
        self._framework = None
        self._trace = None
        self._tags = tags

    def on_predict_begin(self, logs=None):
        self._configure()

    def on_predict_end(self, logs=None):
        graphsignal.upload()

    def on_predict_batch_begin(self, batch, logs=None):
        self._start_trace('predict_batch', batch)

    def on_predict_batch_end(self, batch, logs=None):
        self._stop_trace()

    def _configure(self):
        try:
            if not self._is_initialized:
                self._is_initialized = True

                self._framework = signals_pb2.FrameworkInfo()
                self._framework.type = signals_pb2.FrameworkInfo.FrameworkType.KERAS_FRAMEWORK
                parse_semver(self._framework.version, keras.__version__)
        except Exception:
            logger.error('Error configuring Keras callback', exc_info=True)

    def _start_trace(self, endpoint, batch_idx):
        if not self._trace:
            self._trace = graphsignal.start_trace(
                endpoint=endpoint,
                tags=self._tags)
            self._trace.set_tag('batch', batch_idx)

    def _stop_trace(self):
        if self._trace:
            if self._trace.is_sampling():
                self._update_signal()
            self._trace.stop()
            self._trace = None

    def _update_signal(self):
        try:
            signal = self._trace._signal

            if self._framework:
                signal.frameworks.append(self._framework)
        except Exception as exc:
            logger.debug('Error in Hugging Face callback', exc_info=True)
            self._trace._add_agent_exception(exc)
